/*
 * Direct3D9 textures
 *  Copyright (c) 2003, TransGaming Technologies Inc.
*/

#include "config.h"

#include "winbase.h"
#include "d3d9.h"
#include "wine/debug.h"

#include "d3d9_private.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

static D3D_Light default_light = {
  {D3DLIGHT_DIRECTIONAL,
   {1,1,1,0},
   {0,0,0,0},
   {0,0,0,0},
   {0,0,0},
   {0,0,1},
   0,0,0,0,0,0,0},
  0, 0, {0,0,0,0}
};

/* table to convert from D3D9 world transform to D3D7 world transform */
static UINT wxfrm[4] = {1,4,5,6};

HRESULT WINAPI Direct3DDevice9_SetTransform(LPDIRECT3DDEVICE9 iface,
					    D3DTRANSFORMSTATETYPE dtstType,
					    CONST D3DMATRIX* lpD3DMatrix)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_SETTRANSFORM, 0, {1} };
  D3DHAL_DP2SETTRANSFORM xfrm;

  TRACE("(%p)->(%d,%p)\n", This, dtstType, lpD3DMatrix);
  if (!dtstType) return D3DERR_INVALIDCALL;

  if (dtstType >= 256) {
    unsigned i = dtstType - 256;
    if (i >= 4) return D3DERR_INVALIDCALL;
    dtstType = wxfrm[dtstType-256];
  }

  if (dtstType >= D3DTS_TEXTURE0) {
    unsigned i = dtstType - D3DTS_TEXTURE0;
    if (i >= 8) return D3DERR_INVALIDCALL;
    This->d3dp.txfrm[i] = *lpD3DMatrix;
  }
  else if (dtstType < 8) {
    This->d3dp.xfrm[dtstType] = *lpD3DMatrix;
  }
  xfrm.xfrmType = dtstType;
  xfrm.matrix = *lpD3DMatrix;
  return This->emit_dp2(This->ip_ctx, &cmd, &xfrm, sizeof(xfrm), NULL, 0);
}

HRESULT WINAPI Direct3DDevice9_GetTransform(LPDIRECT3DDEVICE9 iface,
					    D3DTRANSFORMSTATETYPE dtstType,
					    D3DMATRIX* lpD3DMatrix)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  TRACE("(%p)->(%d,%p)\n", This, dtstType, lpD3DMatrix);
  if (!dtstType) return D3DERR_INVALIDCALL;

  if (dtstType >= 256) {
    unsigned i = dtstType - 256;
    if (i >= 4) return D3DERR_INVALIDCALL;
    dtstType = wxfrm[dtstType-256];
  }

  if (dtstType >= D3DTS_TEXTURE0) {
    unsigned i = dtstType - D3DTS_TEXTURE0;
    if (i >= 8) return D3DERR_INVALIDCALL;
    *lpD3DMatrix = This->d3dp.txfrm[i];
  }
  else if (dtstType < 8) {
    *lpD3DMatrix = This->d3dp.xfrm[dtstType];
  }
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_SetViewport(LPDIRECT3DDEVICE9 iface,
					   CONST D3DVIEWPORT9* lpData)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3DHAL_DP2COMMAND cmd = { 0, 0, {1} };
  D3DHAL_DP2VIEWPORTINFO view;
  D3DHAL_DP2ZRANGE zrange;
  HRESULT hr;

  TRACE("(%p)->(%p)\n", This, lpData);
  memcpy(&This->d3dp.viewport, lpData, sizeof(D3DVIEWPORT7));
  view.dwX = lpData->X;
  view.dwY = lpData->Y;
  view.dwWidth  = lpData->Width;
  view.dwHeight = lpData->Height;
  zrange.dvMinZ = lpData->MinZ;
  zrange.dvMaxZ = lpData->MaxZ;
  cmd.bCommand = D3DDP2OP_VIEWPORTINFO;
  hr = This->emit_dp2(This->ip_ctx, &cmd, &view, sizeof(view), NULL, 0);
  /* FIXME: do we also need to do anything about WINFO? */
  if (SUCCEEDED(hr)) {
      cmd.bCommand = D3DDP2OP_ZRANGE;
      hr = This->emit_dp2(This->ip_ctx, &cmd, &zrange, sizeof(zrange), NULL, 0);
  }
  return hr;
}

HRESULT WINAPI Direct3DDevice9_GetViewport(LPDIRECT3DDEVICE9 iface,
					   D3DVIEWPORT9* lpData)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  TRACE("(%p)->(%p)\n", This, lpData);
  memcpy(lpData, &This->d3dp.viewport, sizeof(D3DVIEWPORT7));
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_SetMaterial(LPDIRECT3DDEVICE9 iface,
					   CONST D3DMATERIAL9* lpMat)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_SETMATERIAL, 0, {1} };

  TRACE("(%p)->(%p)\n", This, lpMat);
#if 0
  _dump_colorvalue("Diffuse", lpMat->u.diffuse);
  _dump_colorvalue("Ambient", lpMat->u1.ambient);
  _dump_colorvalue("Specular", lpMat->u2.specular);
  _dump_colorvalue("Emissive", lpMat->u3.emissive);
  TRACE(" Power : %f\n", lpMat->u4.power);
#endif
  return This->emit_dp2(This->ip_ctx, &cmd,
			(D3DMATERIAL7*)lpMat, sizeof(This->d3dp.material), NULL, 0);
}

HRESULT WINAPI Direct3DDevice9_GetMaterial(LPDIRECT3DDEVICE9 iface,
					   D3DMATERIAL9* lpMat)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  TRACE("(%p)->(%p)\n", This, lpMat);
  *(D3DMATERIAL7*)lpMat = This->d3dp.material;
  return D3D_OK;
}

static HRESULT CheckExpandLights(IDirect3DDevice9Impl *This, DWORD dwIndex)
{
  if (dwIndex >= This->d3dp.numlights) {
    DWORD newlights = dwIndex+1, cnt;
    D3DHAL_DP2COMMAND cmd = { D3DDP2OP_CREATELIGHT, 0, {1} };
    D3DHAL_DP2CREATELIGHT data;
    D3D_Light *newlight;
    HRESULT hr;

    newlight = HeapReAlloc(GetProcessHeap(),0,This->d3dp.light,sizeof(D3D_Light)*newlights);
    if (!newlight) {
      ERR("out of memory\n");
      return DDERR_OUTOFMEMORY;
    }
    for (cnt=This->d3dp.numlights; cnt<newlights; cnt++)
      memcpy(&newlight[cnt], &default_light, sizeof(D3D_Light));
    data.dwIndex = dwIndex;
    hr = This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
    if (FAILED(hr)) {
      ERR("CREATELIGHT failed (%ld)\n", hr);
      return hr;
    }
    This->d3dp.light = newlight;
    This->d3dp.numlights = newlights;
  }
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_SetLight(LPDIRECT3DDEVICE9 iface,
					DWORD dwIndex,
					CONST D3DLIGHT9* lpLight)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_SETLIGHT, 0, {1} };
  D3DHAL_DP2SETLIGHT data;
  HRESULT hr;

  TRACE("(%p)->(%ld,%p)\n", This, dwIndex, lpLight);
  hr = CheckExpandLights(This, dwIndex);
  if (FAILED(hr)) return hr;
  This->d3dp.light[dwIndex].l = *(D3DLIGHT7*)lpLight;
  data.dwIndex = dwIndex;
  data.dwDataType = D3DHAL_SETLIGHT_DATA;
  return This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data),
			&This->d3dp.light[dwIndex].l, sizeof(D3DLIGHT7));
}

HRESULT WINAPI Direct3DDevice9_GetLight(LPDIRECT3DDEVICE9 iface,
					DWORD dwIndex,
					D3DLIGHT9* lpLight)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  TRACE("(%p)->(%ld,%p)\n", This, dwIndex, lpLight);
  if (dwIndex < This->d3dp.numlights)
    *(D3DLIGHT7*)lpLight = This->d3dp.light[dwIndex].l;
  else
    *(D3DLIGHT7*)lpLight = default_light.l;
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_LightEnable(LPDIRECT3DDEVICE9 iface,
					   DWORD dwIndex,
					   BOOL bEnable)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_SETLIGHT, 0, {1} };
  D3DHAL_DP2SETLIGHT data;
  DWORD oldflags;
  HRESULT hr;

  TRACE("(%p)->(%ld,%d)\n", This, dwIndex, bEnable);
  hr = CheckExpandLights(This, dwIndex);
  if (FAILED(hr)) return hr;
  oldflags = This->d3dp.light[dwIndex].flags;
  if (bEnable) {
    This->d3dp.light[dwIndex].flags |= D3DLIGHT_ACTIVE;
    data.dwDataType = D3DHAL_SETLIGHT_ENABLE;
  } else {
    This->d3dp.light[dwIndex].flags &= ~D3DLIGHT_ACTIVE;
    data.dwDataType = D3DHAL_SETLIGHT_DISABLE;
  }
  data.dwIndex = dwIndex;
  if (This->d3dp.light[dwIndex].flags == oldflags) return D3D_OK;
  return This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
}

HRESULT WINAPI Direct3DDevice9_GetLightEnable(LPDIRECT3DDEVICE9 iface,
					      DWORD dwIndex,
					      LPBOOL lpbEnable)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  TRACE("(%p)->(%ld,%p)\n", This, dwIndex, lpbEnable);
  if (dwIndex < This->d3dp.numlights)
    *lpbEnable = This->d3dp.light[dwIndex].flags & D3DLIGHT_ACTIVE;
  else
    *lpbEnable = FALSE;
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_SetClipPlane(LPDIRECT3DDEVICE9 iface,
					    DWORD dwIndex,
					    CONST float* pPlane)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_RENDERSTATE, 0, {1} };
  D3DHAL_DP2SETCLIPPLANE data;

  TRACE("(%p)->(%ld,%p)\n", This, dwIndex, pPlane);
  memcpy(&This->d3dp.clipplane[dwIndex], pPlane, sizeof(float)*4);
  data.dwIndex = dwIndex;
  memcpy(&data.plane, pPlane, sizeof(float)*4);
  return This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
}

HRESULT WINAPI Direct3DDevice9_GetClipPlane(LPDIRECT3DDEVICE9 iface,
					    DWORD dwIndex,
					    float* pPlane)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  TRACE("(%p)->(%ld,%p)\n", This, dwIndex, pPlane);
  memcpy(pPlane, &This->d3dp.clipplane[dwIndex], sizeof(float)*4);
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_SetRenderState(LPDIRECT3DDEVICE9 iface,
					      D3DRENDERSTATETYPE d3dType,
					      DWORD dwState)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_RENDERSTATE, 0, {1} };
  D3DHAL_DP2RENDERSTATE stat;

  TRACE("(%p)->(%d,%08lx)\n", This, d3dType, dwState);
  if (d3dType >= 256) {
    WARN("out-of-range render state (%d)\n", d3dType);
    return D3DERR_INVALIDCALL;
  }
  /* according to the HAL docs, it seems the driver is
   * responsible for updating the render state array
   * to what it's capable of handling */
  stat.RenderState = d3dType;
  stat.u.dwState = dwState;
  return This->emit_dp2(This->ip_ctx, &cmd, &stat, sizeof(stat), NULL, 0);
}

HRESULT WINAPI Direct3DDevice9_GetRenderState(LPDIRECT3DDEVICE9 iface,
					      D3DRENDERSTATETYPE d3dType,
					      LPDWORD lpdwState)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  if (d3dType >= 256) {
    WARN("out-of-range render state (%d)\n", d3dType);
    return D3DERR_INVALIDCALL;
  }
  *lpdwState = This->d3dp.renderstate[d3dType];
  TRACE("(%p)->(%d,%p) => %08lx\n", This, d3dType, lpdwState, *lpdwState);
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_BeginStateBlock(LPDIRECT3DDEVICE9 iface)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3D9StateBlock* block;
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_STATESET, 0, {1} };
  D3DHAL_DP2STATESET data;
  HRESULT hr;

  TRACE("(%p)\n", This);
  if (This->current->state_block) return D3DERR_INVALIDCALL;

  block = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(D3D9StateBlock));

  hr = This->alloc_state_block(This->ip_ctx, &block->state_block);
  if (FAILED(hr)) {
    HeapFree(GetProcessHeap(), 0, block);
    return hr;
  }

  data.dwOperation = D3DHAL_STATESETBEGIN;
  data.dwParam = block->state_block;
  data.sbType = D3DSBT_ALL;
  hr = This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
  if (SUCCEEDED(hr)) This->current = block;
  return hr;
}

HRESULT WINAPI Direct3DDevice9_EndStateBlock(LPDIRECT3DDEVICE9 iface,
					     IDirect3DStateBlock9** ppSB)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3D9StateBlock* block;
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_STATESET, 0, {1} };
  D3DHAL_DP2STATESET data;

  TRACE("(%p)->(%p)\n", This, ppSB);
  if (!This->current->state_block) return D3DERR_INVALIDCALL;

  block = This->current;
  *ppSB = (LPDIRECT3DSTATEBLOCK9)block;
  This->current = &This->state;
  TRACE(" => %p\n", *ppSB);

  data.dwOperation = D3DHAL_STATESETEND;
  data.dwParam = block->state_block;
  data.sbType = D3DSBT_ALL;
  return This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
}

HRESULT WINAPI Direct3DDevice9_ApplyStateBlock(LPDIRECT3DDEVICE9 iface,
					       DWORD dwBlock)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3D9StateBlock* block;
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_STATESET, 0, {1} };
  D3DHAL_DP2STATESET data;
  HRESULT hr;

  TRACE("(%p)->(%ld)\n", This, dwBlock);
  if (This->current->state_block) return D3DERR_INVALIDCALL;

  block = (D3D9StateBlock*)dwBlock;

  data.dwOperation = D3DHAL_STATESETEXECUTE;
  data.dwParam = block->state_block;
  data.sbType = D3DSBT_ALL;
  hr = This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
  if (SUCCEEDED(hr)) {
    /* update current state to match new device state */
    unsigned u;
    for (u=0; u<8; u++) {
      if (This->state.tex[u] != block->tex[u]) {
        if (block->tex[u]) IDirect3DBaseTexture9_AddRef(block->tex[u]);
        if (This->state.tex[u]) IDirect3DBaseTexture9_Release(This->state.tex[u]);
        This->state.tex[u] = block->tex[u];
      }
    }
    This->dirty_textures = TRUE; /* just in case */
    /* we could also do states that the driver doesn't know about here */
  }
  return hr;
}

HRESULT WINAPI Direct3DDevice9_CaptureStateBlock(LPDIRECT3DDEVICE9 iface,
						 DWORD dwBlock)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3D9StateBlock* block;
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_STATESET, 0, {1} };
  D3DHAL_DP2STATESET data;
  HRESULT hr;

  TRACE("(%p)->(%ld)\n", This, dwBlock);
  if (This->current->state_block) return D3DERR_INVALIDCALL;

  block = (D3D9StateBlock*)dwBlock;

  data.dwOperation = D3DHAL_STATESETCAPTURE;
  data.dwParam = block->state_block;
  data.sbType = D3DSBT_ALL;
  hr = This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
  if (SUCCEEDED(hr)) {
    /* update state block to match current state */
    unsigned u;
    for (u=0; u<8; u++) {
      if (This->state.tex[u] != block->tex[u]) {
        if (This->state.tex[u]) IDirect3DBaseTexture9_AddRef(This->state.tex[u]);
        if (block->tex[u]) IDirect3DBaseTexture9_Release(block->tex[u]);
        block->tex[u] = This->state.tex[u];
      }
    }
    /* we could also do states that the driver doesn't know about here */
  }
  return hr;
}

HRESULT WINAPI Direct3DDevice9_DeleteStateBlock(LPDIRECT3DDEVICE9 iface,
						DWORD dwBlock)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3D9StateBlock* block;
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_STATESET, 0, {1} };
  D3DHAL_DP2STATESET data;
  HRESULT hr;

  TRACE("(%p)->(%ld)\n", This, dwBlock);

  block = (D3D9StateBlock*)dwBlock;

  data.dwOperation = D3DHAL_STATESETDELETE;
  data.dwParam = block->state_block;
  data.sbType = D3DSBT_ALL;
  hr = This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
  if (FAILED(hr)) return hr;
  hr = This->free_state_block(This->ip_ctx, dwBlock);
  if (SUCCEEDED(hr)) {
    /* release referenced objects */
    unsigned u;
    for (u=0; u<8; u++)
      if (block->tex[u]) IDirect3DBaseTexture9_Release(block->tex[u]);
    HeapFree(GetProcessHeap(), 0, block);
  }
  return hr;
}

HRESULT WINAPI Direct3DDevice9_GetTexture(LPDIRECT3DDEVICE9 iface,
					  DWORD dwStage,
					  LPDIRECT3DBASETEXTURE9* lplpTexture)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  *lplpTexture = This->state.tex[dwStage];
  TRACE("(%p)->(%ld,%p) => %p\n", This, dwStage, lplpTexture, *lplpTexture);
  if (*lplpTexture) IDirect3DBaseTexture9_AddRef(*lplpTexture);
  return D3D_OK;
}

ULONG WINAPI IDirect3DResource9_AddRefInternal(LPDIRECT3DRESOURCE9 iface)
{
  ICOM_THIS(IDirect3DResource9Impl, iface);

  TRACE("(%p)->()\n", iface);
  This->intref++;
  return IDirect3DResource9_AddRef(iface);
}

ULONG WINAPI IDirect3DResource9_ReleaseInternal(LPDIRECT3DRESOURCE9 iface)
{ 
  ICOM_THIS(IDirect3DResource9Impl, iface);

  TRACE("(%p)->()\n", iface);
  This->intref--;
  return IDirect3DResource9_Release(iface);
}
  
HRESULT WINAPI Direct3DDevice9_SetTexture(LPDIRECT3DDEVICE9 iface,
					  DWORD dwStage,
					  LPDIRECT3DBASETEXTURE9 lpTexture)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  LPDIRECT3DBASETEXTURE9 oldtex;
  IDirect3DBaseTexture9Impl* tex = (IDirect3DBaseTexture9Impl*)lpTexture;
  /* in reality, Direct3D allocates "surface handles" that it
   * uses for the state here, but since we don't have the memory
   * restrictions that real drivers do (and we're very lazy too),
   * we'll just cast real pointers here, at least for now */
  DWORD dwState = 0;
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_TEXTURESTAGESTATE, 0, {1} };
  D3DHAL_DP2TEXTURESTAGESTATE stat;

  TRACE("(%p)->(%ld,%p)\n", This, dwStage, lpTexture);

  oldtex = This->current->tex[dwStage];
  if (lpTexture == oldtex) return D3D_OK;

  /* maintain reference count - but the reference count should be hidden 
   * from the app (as per testing on NT2k DX8.1) so an internal refcount
   * is used */
  if (lpTexture) IDirect3DResource9_AddRefInternal((LPDIRECT3DRESOURCE9)lpTexture);
  This->current->tex[dwStage] = lpTexture;
  if (oldtex) IDirect3DResource9_ReleaseInternal((LPDIRECT3DRESOURCE9)oldtex);

  /* find texture object's first surface */
  if (lpTexture)
    dwState = TEXTURE_HANDLE(tex);

  if (!This->current->state_block) {
    This->d3dp.texstagestate[dwStage][D3DTSS_TEXTUREMAP] = dwState;
    if (lpTexture && tex->IsDirty) This->dirty_textures = TRUE;
    TRACE("saving\n");
  }
  TRACE("val=%08lx\n", dwState);
  stat.wStage = dwStage;
  stat.TSState = D3DTSS_TEXTUREMAP;
  stat.dwValue = dwState;
  return This->emit_dp2(This->ip_ctx, &cmd, &stat, sizeof(stat), NULL, 0);
}

HRESULT WINAPI Direct3DDevice9_GetTextureStageState(LPDIRECT3DDEVICE9 iface,
						    DWORD dwStage,
						    D3DTEXTURESTAGESTATETYPE d3dType,
						    LPDWORD lpdwState)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  DWORD dwState;

  if (d3dType >= 64) {
    WARN("out-of-range texture stage state %d\n", d3dType);
    return D3DERR_INVALIDCALL;
  }
  dwState = This->d3dp.texstagestate[dwStage][d3dType];

  TRACE("(%p)->(%ld,%d,%p) => %ld\n", This, dwStage, d3dType, lpdwState, dwState);
  *lpdwState = dwState;

  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_SetTextureStageState(LPDIRECT3DDEVICE9 iface,
						    DWORD dwStage,
						    D3DTEXTURESTAGESTATETYPE d3dType,
						    DWORD dwState)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_TEXTURESTAGESTATE, 0, {1} };
  D3DHAL_DP2TEXTURESTAGESTATE stat;

  TRACE("(%p)->(%ld,%d,%ld)\n", This, dwStage, d3dType, dwState);

  if (d3dType >= 64) {
    WARN("out-of-range texture stage state %d\n", d3dType);
    return D3DERR_INVALIDCALL;
  }

  /* the doc implies that the app is allowed to set any state it wants,
   * _then_ check (with ValidateDevice) whether the hardware groks the
   * resulting state vector, so I guess no checks should be performed here
   */
  if (!This->current->state_block)
    This->d3dp.texstagestate[dwStage][d3dType] = dwState;
  stat.wStage = dwStage;
  stat.TSState = d3dType;
  stat.dwValue = dwState;
  return This->emit_dp2(This->ip_ctx, &cmd, &stat, sizeof(stat), NULL, 0);
}

static const D3DTEXTURESTAGESTATETYPE sampler_type[16] = {
  0,
  13, /* ADDRESSU */
  14, /* ADDRESSV */
  25, /* ADDRESSW */
  15, /* BORDERCOLOR */
  16, /* MAGFILTER */
  17, /* MINFILTER */
  18, /* MIPFILTER */
  19, /* MIPMAPLODBIAS */
  20, /* MAXMIPLEVEL */
  21  /* MAXANISOTROPY */
  /* FIXME: SRGBTEXTURE, ELEMENTINDEX, DMAPOFFSET */
};

HRESULT WINAPI Direct3DDevice9_GetSamplerState(LPDIRECT3DDEVICE9 iface,
					       DWORD Sampler,
					       D3DSAMPLERSTATETYPE Type,
					       DWORD* pValue)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3DTEXTURESTAGESTATETYPE d3dType = sampler_type[Type];
  DWORD Value;

  /* tables to convert from D3D7 texture filters to D3D9 texture filters */
  static const DWORD magfilter[6] = {0, 1, 2, 6, 7, 3};
  static const DWORD minfilter[4] = {0, 1, 2, 3};
  static const DWORD mipfilter[4] = {0, 0, 1, 2};

  if (Type >= 16) {
    WARN("out-of-range sampler type %d\n", Type);
    return D3DERR_INVALIDCALL;
  }

  if (!d3dType && Type) {
    WARN("unsupported sampler type %d\n", Type);
    return D3DERR_INVALIDCALL;
  }

  Value = This->d3dp.texstagestate[Sampler][d3dType];

  /* convert from D3D7 (HAL) texture filters to D3D9 texture filters */
  switch (Type) {
  case D3DSAMP_MAGFILTER:
    Value = magfilter[Value];
    break;
  case D3DSAMP_MINFILTER:
    Value = minfilter[Value];
    break;
  case D3DSAMP_MIPFILTER:
    Value = mipfilter[Value];
    break;
  default:
    break;
  }

  TRACE("(%p)->(%ld,%d,%p) => %ld\n", This, Sampler, Type, pValue, Value);
  *pValue = Value;

  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_SetSamplerState(LPDIRECT3DDEVICE9 iface,
					       DWORD Sampler,
					       D3DSAMPLERSTATETYPE Type,
					       DWORD Value)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_TEXTURESTAGESTATE, 0, {1} };
  D3DHAL_DP2TEXTURESTAGESTATE stat;
  D3DTEXTURESTAGESTATETYPE d3dType = sampler_type[Type];

  /* tables to convert from D3D9 texture filters to D3D7 texture filters */
  static const DWORD magfilter[8] = {0, 1, 2, 5, 3, 4, 3, 4};
  static const DWORD minfilter[8] = {0, 1, 2, 3, 3, 3, 3, 3};
  static const DWORD mipfilter[8] = {1, 2, 3, 3, 3, 3, 3, 3};

  TRACE("(%p)->(%ld,%d,%ld)\n", This, Sampler, Type, Value);

  if (Type >= 16) {
    WARN("out-of-range sampler type %d\n", Type);
    return D3DERR_INVALIDCALL;
  }

  if (!d3dType && Type) {
    WARN("unsupported sampler type %d\n", Type);
    return D3DERR_INVALIDCALL;
  }

  /* convert from D3D9 texture filters to D3D7 (HAL) texture filters */
  switch (Type) {
  case D3DSAMP_MAGFILTER:
    Value = magfilter[Value];
    break;
  case D3DSAMP_MINFILTER:
    Value = minfilter[Value];
    break;
  case D3DSAMP_MIPFILTER:
    Value = mipfilter[Value];
    break;
  default:
    break;
  }

  if (!This->current->state_block)
    This->d3dp.texstagestate[Sampler][d3dType] = Value;
  stat.wStage = Sampler;
  stat.TSState = d3dType;
  stat.dwValue = Value;
  return This->emit_dp2(This->ip_ctx, &cmd, &stat, sizeof(stat), NULL, 0);
}

HRESULT WINAPI Direct3DDevice9_ValidateDevice(LPDIRECT3DDEVICE9 iface,
					      DWORD* pNumPasses)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  TRACE("(%p)->(%p)\n", This, pNumPasses);
  return This->validate_device(This->ip_ctx, pNumPasses);
}

HRESULT WINAPI Direct3DDevice9_SetStreamSource(LPDIRECT3DDEVICE9 iface,
					       UINT StreamNumber,
					       LPDIRECT3DVERTEXBUFFER9 pStreamData,
					       UINT OffsetInBytes,
					       UINT Stride)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  LPDIRECT3DVERTEXBUFFER9 olddata;
  IDirect3DVertexBuffer9Impl* buffer = (IDirect3DVertexBuffer9Impl *)pStreamData;

  TRACE("(%p)->(%d,%p,%d)\n", This, StreamNumber, pStreamData, Stride);

  if (OffsetInBytes) {
    FIXME("offset=%d\n", OffsetInBytes);
    return D3DERR_INVALIDCALL;
  }

  olddata = This->streamsource[StreamNumber];

  if (pStreamData != olddata) {
    if (pStreamData) IDirect3DResource9_AddRefInternal((LPDIRECT3DRESOURCE9)pStreamData);
    This->streamsource[StreamNumber] = pStreamData;
    if (olddata) IDirect3DResource9_ReleaseInternal((LPDIRECT3DRESOURCE9)olddata);
  }

  This->d3dp.stream[StreamNumber].data_t = &(buffer->t);
  This->d3dp.stream[StreamNumber].stride = Stride;

  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_GetStreamSource(LPDIRECT3DDEVICE9 iface,
					       UINT StreamNumber,
					       LPDIRECT3DVERTEXBUFFER9* ppStreamData,
					       UINT* pOffsetInBytes,
					       UINT* pStride)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  *ppStreamData = This->streamsource[StreamNumber];
  *pOffsetInBytes = 0;
  *pStride = This->d3dp.stream[StreamNumber].stride;
  TRACE("(%p)->(%d,%p,%p)\n", This, StreamNumber, ppStreamData, pStride);
  if (*ppStreamData) IDirect3DVertexBuffer9_AddRef(*ppStreamData);
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_SetIndices(LPDIRECT3DDEVICE9 iface,
					  LPDIRECT3DINDEXBUFFER9 pIndexData)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  LPDIRECT3DINDEXBUFFER9 olddata;

  TRACE("(%p)->(%p)\n", This, pIndexData);

  olddata = This->index_data;

  if (pIndexData != olddata) {
    if (pIndexData) IDirect3DResource9_AddRefInternal((LPDIRECT3DRESOURCE9)pIndexData);
    This->index_data = pIndexData;
    if (olddata) IDirect3DResource9_ReleaseInternal((LPDIRECT3DRESOURCE9)olddata);
  }

  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_GetIndices(LPDIRECT3DDEVICE9 iface,
					  LPDIRECT3DINDEXBUFFER9* ppIndexData)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  *ppIndexData = This->index_data;
  TRACE("(%p)->(%p)\n", This, ppIndexData);
  if (*ppIndexData) IDirect3DIndexBuffer9_AddRef(*ppIndexData);
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_SetFVF(LPDIRECT3DDEVICE9 iface,
                                      DWORD FVF)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface); 
  TRACE("(%p)->(0x%08lx)\n", This, FVF);

  /* TODO: do we need to NULL out the vertex shader here or
   * is SetVertexShader(NULL) guaranteed to be called before this? */
  This->d3dp.vertex_fvf = FVF;
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice9_GetFVF(LPDIRECT3DDEVICE9 iface,
                                      DWORD *FVF)
{
  ICOM_THIS(IDirect3DDevice9Impl, iface);
  TRACE("(%p)->(%p)\n", This, FVF);

  /* retrieve the FVF from where SetFVF put it and assign to *FVF */
  *FVF = This->d3dp.vertex_fvf;
  return D3D_OK;
}
